﻿namespace HeuristicLab.Problems.ProgramSynthesis.Push.Expressions {
  using System;
  using System.Collections.Generic;
  using System.Linq;

  using Attributes;
  using Data.Pool;
  using Stack;

  public static class ExpressionTable {
    public static readonly IReadOnlyDictionary<Type, Expression> StatelessExpressionTable;
    public static readonly IReadOnlyDictionary<Type, Func<Expression>> StatefulExpressionFactory;
    public static readonly IReadOnlyList<string> ExpressionNames;
    public static readonly int ExpressionCount;

    private static readonly Dictionary<int, string> indexToNameTable = new Dictionary<int, string>();
    private static readonly Dictionary<Type, string> typeToNameTable = new Dictionary<Type, string>();
    private static readonly Dictionary<string, Type> nameToTypeTable = new Dictionary<string, Type>();
    private static readonly Dictionary<StackTypes, IList<string>> stackTypeToNamesTable = new Dictionary<StackTypes, IList<string>>();
    private static readonly Dictionary<StackTypes, IList<string>> stackDependencyToNamesTable = new Dictionary<StackTypes, IList<string>>();
    private static readonly Dictionary<Type, PushExpressionAttribute> typeToAttributeTable = new Dictionary<Type, PushExpressionAttribute>();

    public static readonly IReadOnlyList<Type> StatelessExpressionTypes;
    public static readonly IReadOnlyList<Type> StatefulExpressionTypes;

    static ExpressionTable() {
      StatelessExpressionTypes = GetExpressionTypes(typeof(StatelessExpression)).ToArray();
      StatefulExpressionTypes = GetExpressionTypes(typeof(StatefulExpression<>)).ToArray();

      StatelessExpressionTable = GetStatelessExpressionTable();
      StatefulExpressionFactory = GetStatefulExpressionFactory();

      ExpressionNames = StatelessExpressionTable.Keys
        .Concat(StatefulExpressionFactory.Keys)
        .Where(t => typeToNameTable.ContainsKey(t))
        .Select(type => typeToNameTable[type])
        .ToArray();

      ExpressionCount = StatelessExpressionTable.Count + StatefulExpressionFactory.Count;
    }

    public static IReadOnlyDictionary<int, string> IndexToNameTable { get { return indexToNameTable; } }
    public static IReadOnlyDictionary<Type, string> TypeToNameTable { get { return typeToNameTable; } }
    public static IReadOnlyDictionary<string, Type> NameToTypeTable { get { return nameToTypeTable; } }
    public static IReadOnlyDictionary<StackTypes, IList<string>> StackTypeToNamesTable { get { return stackTypeToNamesTable; } }
    public static IReadOnlyDictionary<StackTypes, IList<string>> StackDependencyToNamesTable { get { return stackDependencyToNamesTable; } }
    public static IReadOnlyDictionary<Type, PushExpressionAttribute> TypeToAttributeTable { get { return typeToAttributeTable; } }

    private static Dictionary<Type, Expression> GetStatelessExpressionTable() {
      var dictionary = new Dictionary<Type, Expression>();

      foreach (var type in StatelessExpressionTypes) {
        var expression = Activator.CreateInstance(type) as Expression;
        var attribute = (PushExpressionAttribute)Attribute.GetCustomAttribute(type, typeof(PushExpressionAttribute));

        typeToAttributeTable.Add(type, attribute);
        dictionary.Add(type, expression);

        // do not index hidden expressions like push expression in tables
        if (attribute.IsHidden) continue;

        indexToNameTable.Add(indexToNameTable.Keys.Count, attribute.Name);
        typeToNameTable.Add(type, attribute.Name);
        nameToTypeTable.Add(attribute.Name, type);

        if (!stackTypeToNamesTable.ContainsKey(attribute.StackType)) {
          stackTypeToNamesTable.Add(attribute.StackType, new List<string>());
        }

        stackTypeToNamesTable[attribute.StackType].Add(attribute.Name);

        var dependencies = attribute.StackType | attribute.AdditionalStackDependencies;
        if (!stackDependencyToNamesTable.ContainsKey(dependencies)) {
          stackDependencyToNamesTable.Add(dependencies, new List<string>());
        }

        stackDependencyToNamesTable[dependencies].Add(attribute.Name);
      }

      return dictionary;
    }

    private static Dictionary<Type, Func<Expression>> GetStatefulExpressionFactory() {
      var dictionary = new Dictionary<Type, Func<Expression>>();

      foreach (var type in StatefulExpressionTypes) {
        // Make a NewExpression that calls the ctor
        var newExp = System.Linq.Expressions.Expression.New(type);

        // Create a lambda with the New expression as body
        var creator = System.Linq.Expressions.Expression.Lambda<Func<Expression>>(newExp).Compile();
        var attribute = (PushExpressionAttribute)Attribute.GetCustomAttribute(type, typeof(PushExpressionAttribute));

        typeToAttributeTable.Add(type, attribute);
        dictionary.Add(type, creator);

        // do not index hidden expressions like push expression in tables
        if (attribute.IsHidden) continue;

        indexToNameTable.Add(indexToNameTable.Keys.Count, attribute.Name);
        typeToNameTable.Add(type, attribute.Name);
        nameToTypeTable.Add(attribute.Name, type);

        if (!stackTypeToNamesTable.ContainsKey(attribute.StackType)) {
          stackTypeToNamesTable.Add(attribute.StackType, new List<string>());
        }

        stackTypeToNamesTable[attribute.StackType].Add(attribute.Name);

        var dependencies = attribute.StackType | attribute.AdditionalStackDependencies;
        if (!stackDependencyToNamesTable.ContainsKey(dependencies)) {
          stackDependencyToNamesTable.Add(dependencies, new List<string>());
        }

        stackDependencyToNamesTable[dependencies].Add(attribute.Name);
      }

      return dictionary;
    }

    private static bool IsSubclassOf(Type type, Type baseType) {
      if (type == null || baseType == null || type == baseType)
        return false;

      if (baseType.IsGenericType == false) {
        if (type.IsGenericType == false)
          return type.IsSubclassOf(baseType);
      } else {
        baseType = baseType.GetGenericTypeDefinition();
      }

      type = type.BaseType;
      var objectType = typeof(object);
      while (type != objectType && type != null) {
        var curentType = type.IsGenericType ? type.GetGenericTypeDefinition() : type;
        if (curentType == baseType)
          return true;

        type = type.BaseType;
      }

      return false;
    }

    private static IEnumerable<Type> GetExpressionTypes(Type baseType) {
      return from domainAssembly in AppDomain.CurrentDomain.GetAssemblies()
             from assemblyType in domainAssembly.GetTypes()
             where !assemblyType.IsAbstract &&
                   IsSubclassOf(assemblyType, baseType) &&
                   assemblyType.GetConstructor(Type.EmptyTypes) != null
             select assemblyType;
    }

    public static IReadOnlyList<string> GetExpressionsByStackTypes(StackTypes allowedTypes) {
      return stackDependencyToNamesTable
          .Where(entry => (entry.Key & ~allowedTypes) == 0x0)
          .SelectMany(entry => entry.Value)
          .ToArray();
    }

    public static PushProgram GetProgram(IManagedPool<PushProgram> pool, int[] index) {
      var expressions = new Expression[index.Length];

      for (var i = 0; i < index.Length; i++) {
        expressions[i] = GetExpression(index[i]);
      }

      return PushProgram.Create(pool, expressions);
    }

    public static Expression GetExpression(int index) {
      return GetExpression(indexToNameTable[index]);
    }

    public static Expression GetExpression(Type type) {
      var name = TypeToNameTable[type];
      return GetExpression(name);
    }

    public static Expression GetExpression(string name) {
      Expression expression;

      if (!TryGetStatelessExpression(name, out expression) &&
          !TryGetStatefulExpression(name, out expression))
        throw new InvalidOperationException(string.Format("Expression with name {0} not found", name));

      return expression;
    }

    public static Expression GetStatelessExpression<T>() where T : StatelessExpression {
      return GetStatelessExpression(typeToNameTable[typeof(T)]);
    }

    public static Expression GetStatelessExpression(string name) {
      if (NameToTypeTable.ContainsKey(name) && StatelessExpressionTable.ContainsKey(NameToTypeTable[name]))
        return StatelessExpressionTable[NameToTypeTable[name]];
      throw new NotSupportedException("Expression not supported: " + name);
    }

    public static bool TryGetStatelessExpression(string name, out Expression expression) {
      if (!NameToTypeTable.ContainsKey(name) || !StatelessExpressionTable.ContainsKey(NameToTypeTable[name])) {
        expression = null;
        return false;
      }

      expression = StatelessExpressionTable[NameToTypeTable[name]];
      return true;
    }

    public static Expression GetStatefulExpression<T>() {
      return GetStatefulExpression(typeToNameTable[typeof(T)]);
    }

    public static Expression GetStatefulExpression(string name) {
      if (NameToTypeTable.ContainsKey(name)) {
        Func<Expression> creator;
        var type = NameToTypeTable[name];

        if (StatefulExpressionFactory.TryGetValue(type, out creator))
          return creator();
      }

      throw new NotSupportedException("Expression not supported: " + name);
    }

    public static bool TryGetStatefulExpression(string name, out Expression expression) {
      if (NameToTypeTable.ContainsKey(name)) {
        expression = StatefulExpressionFactory[NameToTypeTable[name]]();
        return true;
      }

      expression = default(Expression);
      return false;
    }
  }
}